#!/usr/bin/env python

##===--- iwyu-mapgen-cpython.py -------------------------------------------===##
#
#                     The LLVM Compiler Infrastructure
#
# This file is distributed under the University of Illinois Open Source
# License. See LICENSE.TXT for details.
#
##===----------------------------------------------------------------------===##

""" Generate mappings for Python C API headers.

The Python include root in e.g. /usr/include/python3.8 contains a single public
header: Python.h.

Simply collect all included header names and map them all to Python.h.
"""
import os
import re
import sys
import json
import argparse
import fnmatch


INCLUDE_RE = re.compile(r'#\s*include\s+"([^"]+)"')


def parse_include_names(headerpath):
    """ Parse the header file at headerpath and return all include names. """
    with open(headerpath, 'r') as fobj:
        for line in fobj.readlines():
            m = INCLUDE_RE.search(line)
            if m:
                yield m.group(1)


def iterfiles(dirpath, pattern):
    """ Recursively find all files matching pattern. """
    for root, _, files in os.walk(dirpath):
        for fname in files:
            if fnmatch.fnmatch(fname, pattern):
                yield os.path.join(root, fname)


def generate_imp_lines(include_names):
    """ Generate a sequence of json-formatted strings in .imp format.

    This should ideally return a jsonable structure instead, and use json.dump
    to write it to the output file directly. But there doesn't seem to be a
    simple way to convince Python's json library to generate a "packed"
    formatting, it always prefers to wrap dicts onto multiple lines.

    Cheat, and use json.dumps for escaping each line.
    """
    def jsonline(mapping, indent):
        return (indent * ' ') + json.dumps(mapping)

    for name in sorted(include_names):
        # Regex-escape period and build a regex matching both "" and <>.
        map_from = r'@["<]%s[">]' % name.replace('.', '\\.')
        mapping = {'include': [map_from, 'private', '<Python.h>', 'public']}
        yield jsonline(mapping, indent=2)


def main(pythonroot):
    """ Entry point. """

    # Collect all include names in the root. These are the private includes.
    included_names = []
    for fname in iterfiles(pythonroot, '*.h'):
        included_names.extend(parse_include_names(fname))

    # Discard duplicates and remove Python.h itself.
    included_names = set(included_names)
    included_names.remove('Python.h')

    # Print mappings from name -> Python.h.
    print('[')
    print(',\n'.join(generate_imp_lines(sorted(included_names))))
    print(']')

    return 0


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='Generate IWYU mappings for the Python C API.')
    parser.add_argument('pythonroot',
                        help='Python include root (e.g. /usr/include/python3.8')
    args = parser.parse_args()
    sys.exit(main(args.pythonroot))



#!/usr/bin/env python

##===--- run_iwyu_tests.py - include-what-you-use test framework driver ---===##
#
#                     The LLVM Compiler Infrastructure
#
# This file is distributed under the University of Illinois Open Source
# License. See LICENSE.TXT for details.
#
##===----------------------------------------------------------------------===##

"""A test harness for IWYU testing."""

__author__ = 'dsturtevant@google.com (Dean Sturtevant)'

import argparse
import glob
import os
import re
import sys
import unittest
import logging
logging.basicConfig(level=logging.INFO)
import posixpath
from fnmatch import fnmatch
import iwyu_test_util


def PosixPath(path):
    """Normalize Windows path separators to POSIX path separators."""
    return path.replace('\\', '/')


def Partition(l, delimiter):
  try:
    delim_index = l.index(delimiter)
  except ValueError:
    return l, []

  return l[:delim_index], l[delim_index+1:]


def TestIwyuOnRelevantFiles(filename):
  logging.info('Testing iwyu on %s', filename)
  # Split full/path/to/foo.cc into full/path/to/foo and .cc.
  (all_but_extension, extension) = os.path.splitext(filename)
  (dirname, basename) = os.path.split(all_but_extension)
  # Generate diagnostics on all foo-* files (well, not other
  # foo-*.cc files, which is not kosher but is legal), in addition
  # to foo.h (if present) and foo.cc.
  all_files = (glob.glob('%s-*' % all_but_extension) +
               glob.glob('%s/*/%s-*' % (dirname, basename)) +
               glob.glob('%s.h' % all_but_extension) +
               glob.glob('%s/*/%s.h' % (dirname, basename)))
  
  files_to_check = [f for f in all_files if not f.endswith(extension)]
  files_to_check.append(filename)

  # IWYU emits summaries with canonicalized filepaths, where all the
  # directory separators are set to '/'. In order for the testsuite to
  # correctly match up file summaries, we must canonicalize the filepaths
  # in the same way here.
  files_to_check = [PosixPath(f) for f in files_to_check]
  iwyu_test_util.TestIwyuOnRelativeFile(filename, files_to_check, verbose=True)


def GenerateTests(rootdir, pattern):
  def _AddTestFunctions(cls):
    filenames = []
    test_files = {}
    for (dirpath, _, files) in os.walk(rootdir):
      dirpath = PosixPath(dirpath)  # Normalize path separators.
      filenames.extend(posixpath.join(dirpath, f) for f in files
                       if fnmatch(f, pattern))
    if not filenames:
      print('No tests found in %s!' % os.path.abspath(rootdir))
      return

    for filename in filenames:
      all_but_extension = os.path.splitext(filename)[0]
      basename = os.path.basename(all_but_extension)
      test_name = re.sub('[^0-9a-zA-Z_]', '_', basename)  # python-clean
      test_name = 'test_%s' % test_name

      while hasattr(cls, test_name):   # already have a class with that name
        test_name += '2'               # just append a suffix :-)

      setattr(cls, test_name, lambda x, f=filename: TestIwyuOnRelevantFiles(f))
      test_files[test_name] = filename

    setattr(cls, 'test_files', test_files)

    return cls

  return _AddTestFunctions


def EnumerateLoadedTests():
  for suite in unittest.defaultTestLoader.loadTestsFromModule(sys.modules[__name__]):
    for test in suite:
      yield (test.__class__, test._testMethodName)


def PrintLoadedTests():
  for (cls, test) in EnumerateLoadedTests():
    print('%s.%s' % (cls.__name__, test))


def PrintLoadedTestsAndFiles():
  for (cls, test) in EnumerateLoadedTests():
    print('%s.%s:%s' % (cls.__name__, test, cls.test_files[test]))


if __name__ == '__main__':
  unittest_args, additional_args = Partition(sys.argv, '--')
  if additional_args:
    iwyu_test_util.SetIwyuPath(additional_args[0])

  parser = argparse.ArgumentParser(add_help=False, usage=argparse.SUPPRESS)
  group = parser.add_mutually_exclusive_group()
  group.add_argument('--list', dest='list_tests', action='store_true')
  group.add_argument('--list-test-files', action='store_true')
  group.add_argument('--run-test-file')
  (runner_args, _) = parser.parse_known_args(unittest_args)

  if runner_args.run_test_file:
    exit(TestIwyuOnRelevantFiles(runner_args.run_test_file))

  @GenerateTests(rootdir='tests/c', pattern='*.c')
  class c(unittest.TestCase): pass

  @GenerateTests(rootdir='tests/cxx', pattern='*.cc')
  class cxx(unittest.TestCase): pass

  @GenerateTests(rootdir='tests/driver', pattern='*.c')
  class driver(unittest.TestCase): pass

  if runner_args.list_tests:
    exit(PrintLoadedTests())
  elif runner_args.list_test_files:
    exit(PrintLoadedTestsAndFiles())

  unittest.main(argv=unittest_args)



#!/usr/bin/env python

##===--- iwyu-mapgen-qt.py ------------------------------------------------===##
#
#                     The LLVM Compiler Infrastructure
#
# This file is distributed under the University of Illinois Open Source
# License. See LICENSE.TXT for details.
#
##===----------------------------------------------------------------------===##

""" Generates mappings for Qt API headers.

Qt has quite a strong module convention, where there's one public header for
every module class, e.g.:

- For a module X there's typically...
- ... a set of classes called QtXy, QtXyz...
- ... and a corresponding public header for each called QtXy, QtXyz...
- ... and possibly a set of private headers called qtxy.h, qtxyz.h...

Use these conventions to generate symbol and include mappings for the entire Qt
tree.
"""

import argparse
import glob
import json
import os
import re
import sys


OUTFILEHDR = ("# Do not edit! This file was generated by the script %s." %
              os.path.basename(__file__))

QOBJECT_SYMBOLS = [
    "QObjectList",
    "qFindChildren",
    "qobject_cast",
    "QT_NO_NARROWING_CONVERSIONS_IN_CONNECT",
    "Q_CLASSINFO",
    "Q_DISABLE_COPY",
    "Q_DISABLE_COPY_MOVE",
    "Q_DISABLE_MOVE",
    "Q_EMIT",
    "Q_ENUM",
    "Q_ENUM_NS",
    "Q_FLAG",
    "Q_FLAG_NS",
    "Q_GADGET",
    "Q_INTERFACES",
    "Q_INVOKABLE",
    "Q_NAMESPACE",
    "Q_NAMESPACE_EXPORT",
    "Q_OBJECT",
    "Q_PROPERTY",
    "Q_REVISION",
    "Q_SET_OBJECT_NAME",
    "Q_SIGNAL",
    "Q_SIGNALS",
    "Q_SLOT",
    "Q_SLOTS",
    "emit",
    "slots",
    "signals",
    "SIGNAL",
    "SLOT",
]


class QtHeader(object):
    """ Carry data associated with a Qt header """
    def __init__(self, headername):
        self.headername = headername
        self.classname = os.path.basename(headername)
        self.modulename = os.path.basename(os.path.dirname(headername))
        self._private_headers = None

    def get_private_headers(self):
        """ Return a list of headernames included by this header """
        if self._private_headers is None:
            with open(self.headername, 'r') as headerfile:
                included = re.findall(r'#include "(.*)\.h"', headerfile.read())
            self._private_headers = list(included)
        return self._private_headers


def generate_imp_lines(symbols_map, includes_map):
    """ Generate json-formatted strings in .imp format.

    This should ideally return a jsonable structure instead, and use json.dump
    to write it to the output file directly. But there doesn't seem to be a
    simple way to convince Python's json library to generate a "packed"
    formatting, it always prefers to wrap dicts onto multiple lines.

    Cheat, and use json.dumps for escaping and build a string instead.
    """
    def jsonline(mapping, indent):
        return (indent * " ") + json.dumps(mapping)

    for symbol, header in symbols_map:
        map_to = "<" + header + ">"
        yield jsonline({"symbol": [symbol, "private", map_to, "public"]},
                       indent=2)

    for module, include, header in includes_map:
        # Use regex map-from to match both quoted and angled includes and
        # optional directory prefix (e.g. <QtCore/qnamespace.h> is equivalent to
        # "qnamespace.h").
        map_from = r'@["<](%s/)?%s\.h[">]' % (module, include)
        map_to = "<" + header + ">"
        yield jsonline({"include": [map_from, "private", map_to, "public"]},
                       indent=2)


def add_mapping_rules(header, symbols_map, includes_map):
    """ Add symbol and include mappings for a Qt module. """
    symbols_map += [(header.classname, header.classname)]
    for include in header.get_private_headers():
        includes_map += [(header.modulename, include, header.classname)]


def main(qtroot):
    """ Entry point. """
    symbols_map = []
    includes_map = []
    deferred_headers = []

    # Add manual overrides.
    symbols_map += [("qDebug", "QtGlobal")]
    symbols_map += [(symbol, "QObject") for symbol in QOBJECT_SYMBOLS]
    includes_map += [("QtCore", "qnamespace", "Qt")]

    # Collect mapping information from Qt directory tree.
    headers = glob.glob(os.path.join(qtroot, '**/*[!.h]'))
    for header in headers:
        if os.path.isdir(header):
            continue

        header = QtHeader(header)
        if header.classname == "QInternal":
            continue

        if header.classname == header.modulename:
            deferred_headers.append(header)
        else:
            add_mapping_rules(header, symbols_map, includes_map)

    for header in deferred_headers:
        add_mapping_rules(header, symbols_map, includes_map)

    # Print mappings
    print(OUTFILEHDR)
    print("[")
    print(",\n".join(generate_imp_lines(symbols_map, includes_map)))
    print("]")
    return 0


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("qtroot",
                        help="Qt include root (e.g. /usr/include/.../qt5)")
    args = parser.parse_args()
    sys.exit(main(args.qtroot))



#!/usr/bin/env python

##===--- iwyu-check-license-header.py - check license headers -------------===##
#
#                     The LLVM Compiler Infrastructure
#
# This file is distributed under the University of Illinois Open Source
# License. See LICENSE.TXT for details.
#
##===----------------------------------------------------------------------===##

from __future__ import print_function
import sys
import os
import re
import argparse


# This is used in selected functions to calculate the maximum length of filename
# and filler dashes. Not otherwise useful.
EMBELLISHMENTS = '//===---  ---===//'

HDRFORMAT = """
{c}
{c}                     The LLVM Compiler Infrastructure
{c}
{c} This file is distributed under the University of Illinois Open Source
{c} License. See LICENSE.TXT for details.
{c}
{t}===----------------------------------------------------------------------==={t}
"""


def make_hdrformat(one, two):
    """ Materialize HDRFORMAT based on the one and two comment styles.

    The one and two args should be '#' and '##' or '//' and '//', respectively.
    Returns a list of lines.
    """
    r = HDRFORMAT.lstrip().format(c=one, t=two)
    return r.splitlines()


def truncated(filename):
    """ Truncate the filename with ellipsis if too long """
    maxlen = 80 - len(EMBELLISHMENTS)
    trunclen = maxlen - 3  # ...
    if len(filename) > maxlen:
        filename = filename[:trunclen] + '...'
    return filename


def make_license_header(filename, one, two):
    """ Build a valid license header from filename and comment styles.

    The one and two args should be '#' and '##' or '//' and '//', respectively.
    Returns a list of lines.
    """
    assert len(two) == 2
    filename = os.path.basename(filename)
    filename = truncated(filename)

    def dashes():
        c = 80
        c -= len(EMBELLISHMENTS)
        c -= len(filename)
        return '-' * c

    firstline = '%s===--- %s %s===%s' % (two, filename, dashes(), two)
    return [firstline] + make_hdrformat(one, two)


def format_file_error(filename, *lines):
    """ Format an error message from filename and lines """
    lines = list(lines)
    lines[0] = '%s: %s' % (filename, lines[0])
    return os.linesep.join(lines)


def find_license_header(lines, two):
    """ Return an index where the license header begins """
    if not lines:
        return -1

    for i, line in enumerate(lines):
        # Allow leading blank lines and hash-bangs.
        if not line or line.startswith('#!'):
            continue

        # Besides those, the first line should be a license header.
        if line.startswith(two + '==='):
            break

        # If not, this fails the test entirely.
        return -1

    return i


class File(object):
    """ Base class for a source file with a license header

    Do not use directly, instead use File.parse to instantiate a more derived
    class.

    Derived classes must have three class variables:

    * one - One comment char ('#' for Python, '//' for C++)
    * two - Two comment chars ('##' for Python, '//' for C++)
    * pattern - a regex matching the first line in a license header
    """
    def __init__(self, filename):
        with open(filename, 'r') as fd:
            content = fd.read()

        self.lines = list(content.splitlines())
        self.hdrindex = find_license_header(self.lines, self.two)
        self.filename = filename
        self.errors = []

    @classmethod
    def parse(_, filename):
        """ Return an object derived from File to analyze license headers """
        _, ext = os.path.splitext(filename)
        if ext == '.py':
            klass = PythonFile
        elif ext in ('.h', '.c', '.cc'):
            klass = CxxFile
        else:
            return None

        return klass(filename)

    def has_license_header(self):
        """ Return True if a license header has been found """
        return self.hdrindex != -1

    def add_license_header(self):
        """ Add license header to a file that doesn't have one. """
        assert not self.has_license_header()

        # Find insertion point
        for p, line in enumerate(self.lines):
            # Skip past leading blank lines and hash-bangs.
            if line and not line.startswith('#!'):
                break

        # Split the lines around the insertion point
        if self.lines:
            before, after = self.lines[:p], self.lines[p:]
        else:
            before, after = [], []

        # Rebuild the contents with the license header in the middle
        lines = before
        if before and before[-1] != '':
            lines += ['']
        lines += make_license_header(self.filename, self.one, self.two)
        if after and after[0] != '':
            lines += ['']
        lines += after

        # Write back out
        with open(self.filename, 'wb') as fd:
            fd.write('\n'.join(lines))
            fd.write('\n')

    def check_license_header(self):
        """ Check that the header lines follow convention.

        Returns True if everything is OK, otherwise returns False and populates
        self.errors with all found errors.
        """
        if not self.has_license_header():
            self.file_error('No license header found')
            return False

        hdrlines = self.lines[self.hdrindex:self.hdrindex+8]

        # First line has the most structure
        line = hdrlines[0]
        if len(line) != 80:
            self.line_error(
                1, 'Bad header line length (expected: 80, was: %d)' % len(line),
                " Header line: '%s'" % line)

        m = self.pattern.match(line)
        if not m:
            self.line_error(1, 'Bad header line',
                            " Expected: '%s'" % self.pattern.pattern,
                            " Actual:   '%s'" % line)
        else:
            hfilename = truncated(m.group(1))
            xfilename = truncated(os.path.basename(self.filename))
            if hfilename != xfilename:
                self.line_error(1, 'Bad header filename',
                                " Expected:  '%s'" % xfilename,
                                " Actual:    '%s'" % hfilename)

        # The following seven lines always follow the layout of HDRFORMAT.
        hdrformat = make_hdrformat(self.one, self.two)
        hdrlines = hdrlines[1:]
        for lineno, (expected, actual) in enumerate(zip(hdrformat, hdrlines)):
            if expected != actual:
                self.line_error(lineno + 2, 'Bad header line',
                                " Expected: '%s'" % expected,
                                " Actual:   '%s'" % actual)

        return not self.errors

    def file_error(self, *lines):
        """ Log an error for the file """
        self.errors.append(format_file_error(self.filename, *lines))

    def line_error(self, lineno, *lines):
        """ Log an error for a specific line in the file """
        lines = list(lines)
        lines[0] = '%s:%d: %s' % (self.filename, lineno, lines[0])
        self.errors.append(os.linesep.join(lines))


class PythonFile(File):
    """ Python file with license header """
    one = '#'
    two = '##'
    pattern = re.compile(
        r'##===--- ([a-z0-9_.-]+) -[A-Za-z0-9_.,/# -{}]+===##')

    def __init__(self, filename):
        super(PythonFile, self).__init__(filename)


class CxxFile(File):
    """ C++ file with license header """
    one = '//'
    two = '//'
    pattern = re.compile(
        r'//===--- ([a-z0-9_.-]+) -[A-Za-z0-9_.,/# -{}]+ [-* C+]+===//')

    def __init__(self, filename):
        super(CxxFile, self).__init__(filename)


def main(filenames, add_if_missing):
    """ Entry point.

    Checks license header of all filenames provided.
    Returns zero if all license headers are OK, non-zero otherwise.
    """
    errors = []
    for filename in filenames:
        if os.path.isdir(filename):
            continue

        checker = File.parse(filename)
        if not checker:
            # TODO: Consider printing a warning here in verbose mode.
            continue

        if not checker.check_license_header():
            errors.extend(checker.errors)

        if add_if_missing and not checker.has_license_header():
            checker.add_license_header()

    for err in errors:
        print(err)

    return len(errors)


if __name__ == '__main__':
    parser = argparse.ArgumentParser('IWYU license header checker')
    parser.add_argument('filename', nargs='+')
    parser.add_argument('--add', action='store_true')
    args = parser.parse_args()
    sys.exit(main(args.filename, args.add))
